/*fompronc.c
  Gibt Sensorwerte und Lüfterwerte des Fan-o-Matic pro aus
  Kompilieren bspw.: gcc -Wall -W -O3 -o fompronc fompronc.c tooltipps.c menueausgabe.c -lncursesw -pthread
  Programmiert von Johannes Ullmann
  Nick innovatek Forum: neuling  Nick hardwareluxx Forum: neuli                  */

#include <ncursesw/curses.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/signal.h>
#include <sys/ioctl.h>
#include <string.h>
#include <pthread.h>
#include <locale.h>
#include <langinfo.h>
#include <errno.h>
#include "tooltipps.h"
#include "menueausgabe.h"

#define BAUDRATE B115200
#define FOM "/dev/ttyUSB0"
#define _POSIX_SOURCE 1

static int fd = 0; /* Filedescriptor fuer die serielle Schnittstelle */
static pthread_t thread = 0; /* zum Einfangen des frageTasteAb Fadens */
/* vier ncurses fenster */
WINDOW *links = NULL;
WINDOW *mitte = NULL;
WINDOW *rechts = NULL;
WINDOW *unten = NULL;
WINDOW *haupt = NULL;
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER; /* wird zur Synchronisation der Ausgabe verwendet */
static pthread_mutex_t mutex2 = PTHREAD_MUTEX_INITIALIZER; /* wird zur Synchronisation von Variablen verwendet */
static pthread_mutex_t mutex3 = PTHREAD_MUTEX_INITIALIZER; /* wird zur Synchronisation des FOMpro verwendet */
/* Signalmaske wird verwendet fuer die zu blockenden Signale */
static sigset_t mask;
static sigset_t orig_mask; /* Speicherung der Orginalmaske */
float stemp[16] = {0}; /* Temperatur der Sensoren */
float durchfluss[2] = {0}; /* Der Durchfluss in l/min und l/h*/
static struct termios alt; /* alte serielle Porteinstellungen */
unsigned char woBinIch = 0;
static bool datenDa = FALSE; /* Sind Daten angekommen? */
static bool lauf = TRUE; /* Soll Programm weiterlaufen? */
static bool fensterNeu = FALSE; /* muss Fenster angepasst werden? */
static bool einstellungen = FALSE; /* werden gerade Einstellungen vorgenommen? */
static bool weiter = FALSE; /* muss Modus gewechselt werden? */
/* Die Namen der Sensoren, Regler ... siehe Protokoll_FOMpro_Namen.txt */
char namen[78][21] = {{0}};
unsigned char ameld[32] = {0}; /* Alarm Softwaremeldung */
unsigned char amobo[32] = {0}; /* Alarm Mainboard */
unsigned char areli1[32] = {0}; /* Alarm Relais 1 */
unsigned char areli2[32] = {0}; /* Alarm Relais 2 */
unsigned char a12V[32] = {0}; /* Alarm 12 V Schalter */
unsigned char awga[32] = {0}; /* An welchem Geraet angeschlossen */
unsigned char lspan[16] = {0}; /* Spannung der Luefter */
unsigned char lm[16] = {0}; /* Lueftermodus manuell oder automatisch */
unsigned char lvb[16] = {0}; /* Verbrauch der Luefter */
unsigned char lz[16] = {0}; /* Luefter an oder aus */
unsigned char lupm[16] = {0}; /* Umdrehungen pro Minute der Luefter */
unsigned char ipu[16] = {0}; /* Impulse pro Umdrehung */
unsigned char boost[16] = {0}; /* Boostspannung */
unsigned char minu[16] = {0}; /* Minimalspannung */
unsigned char maxu[16] = {0}; /* Maximalspannung */
unsigned char ks[16] = {0}; /* Korrektur Sensor */
unsigned char rin[16] = {0}; /* Regulator Eingabe */
unsigned char rout[16] = {0}; /* Regulator Ausgabe */
unsigned char rmod[16] = {0}; /* Modus der Regulatoren */
unsigned char rInv[16] = {0}; /* Regulator Ausgang umkehren */
unsigned char tempN[16] = {0}; /* Regulator Temperatur Untergrenze */
unsigned char tempH[16] = {0}; /* Regulator Temperatur Obergrenze */
unsigned char rZeit[16] = {0}; /* Regulatorzeit */
unsigned char rProp[16] = {0}; /* Regulator Proportional */
unsigned char rInt[16] = {0}; /* Regulator Integral */
unsigned char rFo[16] = {0}; /* Regulator Fan off */
unsigned char dfm[7] = {0}; /* Durchflussmesser Alarm, Pulse */
/* Temperatur Step Regulator */
unsigned char rs1Temp[16] = {0};
unsigned char rs2Temp[16] = {0};
unsigned char rs3Temp[16] = {0};
/* Spannung Step Regulator*/
unsigned char rs1U[16] = {0};
unsigned char rs2U[16] = {0};
unsigned char rs3U[16] = {0};
/* Speichert welche 8 Sensorkanaele ausgegeben werden sollen */
unsigned char sensorKanal[8] = {0,1,2,3,4,5,6,7};
unsigned char datum[7] = {0}; /* Datum des FOM pro */
/* Speichert welche 4 Luefterkanaele ausgegeben werden sollen */
unsigned char luefterKanal[4] = {0,1,2,3};
unsigned char nts[3] = {0}; /* Die Netzteilspannung 5V 12V und USB */
unsigned char menu = 0; /* Welcher Menueeintrag ist ausgewaehlt? */
static unsigned char aktiv = 0; /* Welcher Modus ist aktiv? */


/* Ein Fehlerfall ist aufgetreten */
static void gingNichtGut(const char* nachricht) {
	/* Porteinstellungen zurueckschreiben */
	tcsetattr(fd,TCSANOW,&alt);
	/* Serielle Schnittstelle wieder schliessen */
	close(fd);
	/* Wenn noch ein zweiter Programmfaden
	 * noch laeuft muss dieser beendet werden */
	if (0 != thread) {
		sigprocmask(SIG_BLOCK,&mask,&orig_mask);
		pthread_mutex_lock(&mutex2);
		lauf = FALSE;
 		pthread_mutex_unlock(&mutex2);
		sigprocmask(SIG_SETMASK,&orig_mask,NULL);
		pthread_join(thread, NULL);
	}
	/* eventuelle Fenster schliessen */
	if (NULL != links) {
		delwin(links);
	}
	if (NULL != mitte) {
		delwin(mitte);
	}
	if (NULL != rechts) {
		delwin(rechts);
	}
	if (NULL != unten) {
		delwin(unten);
	}
	/* curses mode beenden */
	endwin();
	/* Fehlermeldung ausgeben */
	fprintf(stderr, "%s\n", nachricht);
	/* Programm beenden */
	abort();
}

/* Berechnet wie viele automatische Zeilenumbrueche
   das Monitormenue braucht                  */
static int berechneZeilenumbrueche() {
	int links = 0;
	int rechts = 0;
	unsigned char i = 0;
	if (0 == aktiv) {
		/* zweite Zeile links, die wird 4x ausgegeben */
		links += 26 / ((COLS - 0.5) / 2 - 1);
		links *= 4;
		/* erste Zeile links, die wird 4x ausgegeben */
		for (i = 0; i < 4; i++) {
			links += (25 + strlen(namen[luefterKanal[i]])) / ((COLS - 0.5) / 2 - 1);
		}
		/* zweite zeile rechts, die wird 8x ausgegeben */
		rechts += 9 / ((COLS - 0.5) / 2 - 1);
		rechts *= 8;
		/* erste Zeile rechts, die wird 8x ausgegeben */
		for (i = 0; i < 8; i++) {
			rechts += (16 + strlen(namen[sensorKanal[i] + 16])) / ((COLS - 0.5) / 2 - 1);
		}
		links += 33 / ((COLS - 0.5) / 2 - 1); /* dritte Zeile links */
		links += 57 / ((COLS - 0.5) / 2 - 1); /* vierte Zeile links */
		links += 15 / ((COLS - 0.5) / 2 - 1); /* fuenfte Zeile links */
		links += 52 / ((COLS - 0.5) / 2 - 1); /* sechste Zeile links */
		/* rechts braucht ohne Umbrueche drei Zeilen mehr */
		return (links > rechts + 3) ? links : rechts;
	} else if (1 == aktiv) {
		/* links braucht immer mehr Zeilen als rechts */
		links += 30 / ((COLS - 0.5) / 2 - 1);
		links *= 5;
		links += (10 + strlen(namen[0])) / ((COLS - 0.5) / 2 - 1);
		links += 23 / ((COLS - 0.5) / 2 - 1);
	} else if (2 == aktiv) {
		/* links braucht immer mehr Zeilen als rechts */
		links += 28 / ((COLS -0.5) / 2 - 1);
		links *= 3;
		links += (10 + strlen(namen[0])) / ((COLS - 0.5) / 2 - 1);
	} else if (3 == aktiv) {
		/* mitte braucht immer mehr Zeilen als die anderen beiden */
		links += 28 / (COLS / 3 - 1);
		links = links << 1;
		rechts += 29 / (COLS / 3 - 1);
		links += rechts << 1;
		links += 26 / (COLS / 3 - 1);
		links += 25 / (COLS / 3 - 1);
		
	} else if (4 == aktiv) {
		links += 27 / ((COLS - 0.5) / 2 - 1);
		links *= 3;
	}
	return links;
}

/* Curses Modus starten */
static void starteCurses() {
	int tmp = 0;
	int tmp2 = 0;
	const unsigned char foh[5] = {16, 7, 4, 7, 3}; /* Fenster links rechts Hoehe */
	/* curses Modus starten */
	haupt = initscr();
	/* Bildschirm leeren */
	clear();
	/* abfragen, ob Terminal Farben unterstuetzt */
	if (has_colors() == FALSE) {
		gingNichtGut("Das Terminal unterstuetzt keine Farben!\n");
	}
	/* Farbmodus aktivieren */
	start_color();
	/* Farbpaar 1 definieren */
	init_pair(1, COLOR_BLUE, COLOR_WHITE);
	/* Farbpaar 2 definieren */
	init_pair(2, COLOR_BLUE, COLOR_BLACK);
	/* Farbpaar 3 definieren */
	init_pair(3, COLOR_CYAN, COLOR_BLACK);
	/* raw Modus aber Signale weiterleiten! */
	cbreak();
	/* F1, F2 etc ... */
	keypad(stdscr, TRUE);
	/* auf Tasteneingaben warten */
	nodelay(haupt, FALSE);
	/* Keine Ausgabe beim Holen von Zeichen */
	noecho();
	/* Cursor unsichtbar setzen */
	curs_set(0);
	/* Baue Fenster fuer Uebersichtsmodus */
	tmp = berechneZeilenumbrueche();
	tmp2 = 81 / (COLS - 2);
	if (3 == aktiv) {
		links = newwin(foh[aktiv] + tmp, (COLS - 2) / 3, 2 + tmp2, 1);
		mitte = newwin(foh[aktiv] + tmp, (COLS - 2) / 3, 2 + tmp2, (COLS - 2) / 3);
		rechts = newwin(foh[aktiv] + tmp, (COLS - 2) / 3, 2 + tmp2, (COLS - 2) / 3 * 2);
	} else {
		links = newwin(foh[aktiv] + tmp, (COLS - 0.5) / 2 - 1, 2 + tmp2, 1);
		rechts = newwin(foh[aktiv] + tmp, (COLS - 0.5) / 2 - 1, 2 + tmp2, COLS / 2);
	}
	
	unten = newwin(LINES - foh[aktiv] - 5 - tmp - tmp2, COLS - 2, foh[aktiv] + 3 + tmp + tmp2, 1);
}

/* Umrahmung malen */
static void maleRahmen(int hoehe, int breite) {
	int i = 0;
	/* Blau schwarzer Hintergrund */
	attron(COLOR_PAIR(2));
	/* Oben links Ecke */
	mvwaddch(haupt, 0, 0, ACS_ULCORNER);
	/* Oben Linie von links nach rechts */
	for (i = 1; i < breite - 1; i++) {
		waddch(haupt, ACS_HLINE);
	}
	/* Oben rechts Ecke */
	waddch(haupt, ACS_URCORNER);
	/* Links und rechts Linie nach unten */
	for (i = 1; i < hoehe - 1; i++) {
		mvwaddch(haupt, i, 0, ACS_VLINE);
		mvwaddch(haupt, i, breite - 1, ACS_VLINE);
	}
	/* Links unten Ecke */
	mvwaddch(haupt, hoehe - 1, 0, ACS_LLCORNER);
	/* Unten Linie von links nach rechts */
	for (i = 1; i < breite - 1; i++) {
		waddch(haupt, ACS_HLINE);
	}
	/* Rechts unten Ecke */
	waddch(haupt, ACS_LRCORNER);
	/* Blaue Farbe wieder deaktivieren */
	attroff(COLOR_PAIR(2));
	/* Oben mitte Programmnamen ausgeben */ 
	mvwprintw(haupt, 0, breite / 2 - 4, "fompronc");
}

/* Signal Handler FOM oder Prog hat was gesendet */
void signal_handler_IO (int status) {
	/* Ein Ausgabe Signal */
	if (SIGIO == status) {
		datenDa = TRUE;
	/* interupt Signal strg + c */
	} else if(SIGINT == status) {
		lauf = FALSE;
	/* Fenster veraendert Signal */
	} else if(SIGWINCH == status) {
		fensterNeu = TRUE;
	}
}

/* Setzt den Signalhandler */
static void setzeSignalhandler() {
	struct sigaction saio;
	/* Signalhandler installieren */
	saio.sa_handler = signal_handler_IO;
	sigemptyset(&saio.sa_mask);
	saio.sa_flags = 0;
	saio.sa_restorer = NULL;
	sigaction(SIGIO,&saio,NULL);
	sigaction(SIGINT,&saio,NULL);
	sigaction(SIGWINCH,&saio,NULL);
}

/* Verbindung zum FOM aufnehmen */
static void oeffnePort() {
	/* uninitialisiert */
	struct termios neu;

	/* Serielle Schnittstelle oeffnen */
	fd = open(FOM, O_RDWR | O_NOCTTY | O_NDELAY | O_NONBLOCK);
	/* Schauen obs gut gegangen ist */
	if (fd < 0) {
		perror(FOM);
		abort();
	}

	/* Signalhandler installieren */
	setzeSignalhandler();

	 /* Dem SIGIO Signal erlauben diesen Prozess zu stoeren */
	fcntl(fd, F_SETOWN, getpid());
	/* File Descriptor asynchron setzen */
	fcntl(fd, F_SETFL, FASYNC);

	/* Porteinstellungen sichern bzw. termios intitialisieren */
	tcgetattr(fd,&alt);
	tcgetattr(fd,&neu);

	/* Kontrollflags setzen: */
	neu.c_cflag &= ~(CRTSCTS | CSTOPB | CSIZE | PARENB | PARODD | HUPCL);
	neu.c_cflag |= CLOCAL | CREAD | CS8;
	/* An FOM senden raw Modus */
	neu.c_oflag &= ~(OPOST | OLCUC | ONLCR | OCRNL);
	/* Vom FOM empfangen auch raw Modus */
	neu.c_lflag &= ~(ECHO | ECHOE | ICANON | ISIG);
	/* Einstellen wie die ankommenden Daten vom FOM verarbeitet werden */
	neu.c_iflag &= ~(BRKINT | ICRNL | IGNBRK | IGNCR | IMAXBEL | INLCR |
			INPCK | ISTRIP | IUCLC | IXOFF | IXON | PARMRK);
	neu.c_iflag |= IGNPAR;

	/* Baudrate fuer Ein- und Ausgang setzen */
	cfsetospeed(&neu, BAUDRATE);
	cfsetispeed(&neu, BAUDRATE);

	/* Ein und Ausgabebuffer vom FOM loeschen */
	tcflush(fd, TCIOFLUSH);

	/* Die Einstellungen an den File Desciptor uebergeben */
	tcsetattr(fd, TCSANOW, &neu);
}

/* Stelle Erstkontakt zum FOM pro her */
static void testeVerbindung() {
	int g = 0;
	int i = 0;
	int zaehler = 0;
	char buf[128] = {0};
	char buf2[128] = {0};
	const unsigned char vp[5] = {0x55,0xFF,0x05,0x01,0x00};
	unsigned char versuche = 0;
	unsigned char versatz = 0;
	/* 10 Versuche die Firmwareversion abzufragen */
	for (versuche = 1; versuche < 11; versuche++) {
		printf("Teste Verbindung zum FOM pro. Versuch %d\n", versuche);
		/* FOM Firmwareversion abfragen */
		while (-1 == write(fd,vp,5)) {
			if (EINTR != errno) {
				gingNichtGut("Firmwareversion konnte nicht abgefragt werden!");
			}
		};
		/* Auf Antwort vom FOM warten und abfragen,
		 * ob die Firmwareversion auf 3.0 ist */
		while (lauf) {
			if (datenDa) {
				datenDa = FALSE;
				do {
					g = read(fd,buf,128);
				} while (-1 == g && EINTR == errno);
				/* Startbyte der Antwort suchen */
				while (0x55 != buf[versatz] && versatz < g && 0 == zaehler) versatz++;
				if (versatz == g) {
					versatz = 0;
					if (128 == g) {
						continue;
					} else {
						break;
					}
				}
				for (i = versatz; i < g; i++) {
					buf2[zaehler++] = buf[i];
				}
				versatz = 0;
			}
			if (5 < zaehler) {
				if (3 == buf2[4] && 0 == buf2[5]) {
					return;
				} else {
					break;
				}
			}
			usleep(250000);
			if (!datenDa) {
				break;
			}
		}
		zaehler = 0;
	}
	gingNichtGut("Verbindung zum FOM pro konnte nicht aufgebaut werden.");
}

/* Fragt den FOMpro nach den Namen und speichert diese ins globale Array namen */
static void holeNamen() {
	int g = 0;
	int i = 0;
	int zaehler = 0;
	unsigned short versatz = 0;
	char buf[2048] = {0};
	char buf2[2048] = {0};
	const unsigned char hn[5] = {0x55,0x20,0x05,0x15,0x00};
	unsigned char versuche = 0;

	for (versuche = 0; versuche < 10; versuche++) {
		/* sende Namenabfrage an FOMpro */
		while (-1 == write(fd,hn,5)) {
			if (EINTR != errno) {
				gingNichtGut("Namen konnten nicht abgefragt werden!");
			}
		};
		while (lauf) {
			if (datenDa) {
				datenDa = FALSE;
				do {
					g = read(fd,buf,2048);
				} while (-1 == g && EINTR == errno);
				while (0x65 != buf[versatz] && versatz < g && 0 == zaehler) versatz++;
				if (versatz == g) {
					versatz = 0;
					if (2048 == g) {
						continue;
					} else {
						break;
					}
				}
				for (i = versatz; i < g; i++) {
					buf2[zaehler++] = buf[i];
				}
			}
			if (1663 < zaehler) {
				for (i = 0; i < 78;i++) {
					strncpy(namen[i], &buf2[5 + 21 * i], 20);
					namen[i][20] = '\0';
				}
				return;
			}
			usleep(250000);
			if (!datenDa) {
				zaehler = 0;
				break;
			}
		}
	}
}

static void holeKonfigurationsdaten() {
	int g = 0;
	int i = 0;
	int zaehler = 0;
	unsigned short versatz = 0;
	char buf[1024] = {0};
	char buf2[1024] = {0};
	const unsigned char hk[5] = {0x55,0x20,0x05,0x11,0x00};
	unsigned char versuche = 0;
	for (versuche = 0; versuche < 10; versuche++) {
		/* sende Namenabfrage an FOMpro */
		while (-1 == write(fd,hk,5)) {
			if (EINTR != errno) {
				gingNichtGut("Konfigurationsdaten konnten nicht abgefragt werden!");
			}
		};
		/* werte die Antwort vom FOMpro aus */
		while (lauf) {
			if (datenDa) {
				datenDa = FALSE;
				do {
					g = read(fd,buf,1024);
				} while (-1 == g && EINTR == errno);
				while (0x65 != buf[versatz] && versatz < g && 0 == zaehler) versatz++;
				if (versatz == g) {
					versatz = 0;
					if (1024 == g) {
						continue;
					} else {
						break;
					}
				}
				for (i = versatz; i < g; i++) {
					buf2[zaehler++] = buf[i];
				}
			}
			if (725 < zaehler) {
				/*TODO: Konfigurationsdaten abspeichern
				 * siehe: Protokoll_FOMpro_Konfiguration.txt */
				for (i = 0; i < 16; i++) {
					/* Modus der Luefter abspeichern */
					lm[i] = buf2[i * 14 + 7];
					ameld[i] = buf2[i * 14 + 8];
					amobo[i] = buf2[i * 14 + 9];
					areli1[i] = buf2[i * 14 + 10];
					areli2[i] = buf2[i * 14 + 11];
					a12V[i] = buf2[i * 14 + 12];
					ipu[i] = buf2[i * 14 + 13];
					boost[i] = buf2[i * 14 + 14];
					minu[i] = buf2[i * 14 + 15];
					maxu[i] = buf2[i * 14 + 16];
					awga[i] = buf2[i * 14 + 17];
					/* Sensoren */
					ameld[i + 16] = buf2[i * 10 + 231];
					amobo[i + 16] = buf2[i * 10 + 232];
					areli1[i + 16] = buf2[i * 10 + 233];
					areli2[i + 16] = buf2[i * 10 + 234];
					a12V[i + 16] = buf2[i * 10 + 235];
					ks[i] = buf2[i * 10 + 236];
					awga[i + 16] = buf2[i * 10 + 237];
					/* Regulatoren */
					rin[i] = buf2[i * 16 + 390];
					rout[i] = buf2[i * 16 + 391];
					rmod[i] = buf2[i * 16 + 392];
					rInv[i] = buf2[i * 16 + 393];
					tempN[i] = buf2[i * 16 + 394];
					tempH[i] = buf2[i * 16 + 395];
					rZeit[i] = buf2[i * 16 + 396];
					rProp[i] = buf2[i * 16 + 397];
					rInt[i] = buf2[i * 16 + 398];
					rFo[i] = buf2[i * 16 + 399];
					rs1Temp[i] = buf2[i * 16 + 400];
					rs1U[i] = buf2[i * 16 + 401];
					rs2Temp[i] = buf2[i * 16 + 402];
					rs2U[i] = buf2[i * 16 + 403];
					rs3Temp[i] = buf2[i * 16 + 404];
					rs3U[i] = buf2[i * 16 + 405];
				}
				return;
			}
			usleep(250000);
			if (!datenDa) {
				break;
			}
		}
		zaehler = 0;
	}
}

static void holeKonfigurationsdaten2() {
	int g = 0;
	int i = 0;
	int zaehler = 0;
	unsigned short versatz = 0;
	char buf[512] = {0};
	char buf2[512] = {0};
	const unsigned char hk[5] = {0x55,0x20,0x05,0x12,0x00};
	unsigned char versuche = 0;
	for (versuche = 0; versuche < 10; versuche++) {
		/* sende Namenabfrage an FOMpro */
		while (-1 == write(fd,hk,5)) {
			if (EINTR != errno) {
				gingNichtGut("Konfigurationsdaten konnten nicht abgefragt werden!");
			}
		};
		/* werte die Antwort vom FOMpro aus */
		while (lauf) {
			if (datenDa) {
				datenDa = FALSE;
				do {
					g = read(fd,buf,512);
				} while (-1 == g && EINTR == errno);
				while (0x65 != buf[versatz] && versatz < g && 0 == zaehler) versatz++;
				if (versatz == g) {
					versatz = 0;
					if (333 == g) {
						continue;
					} else {
						break;
					}
				}
				for (i = versatz; i < g; i++) {
					buf2[zaehler++] = buf[i];
				}
			}
			if (332 < zaehler) {
				/*TODO: Konfigurationsdaten abspeichern
				 * siehe: Protokoll_FOMpro_Konfiguration2.txt */
				for (i = 0; i < 7; i++) {
					/* Modus der Luefter abspeichern */
					dfm[i] = buf2[i + 283];
				}
				return;
			}
			usleep(250000);
			if (!datenDa) {
				break;
			}
		}
		zaehler = 0;
	}
}

/* Fenster anpassen */
static void passeFensterAn() {
	struct winsize groessen;
	int tmp = 0;
	int tmp2 = 0;
	const unsigned char foh[5] = {16, 7, 4, 7, 3}; /* Fenster links rechts Hoehe */
	if (-1 == ioctl(fileno (stdout), TIOCGWINSZ, &groessen) ) {
		gingNichtGut("Fenstergroesse konnte nicht angepasst werden");
	}
	keypad(stdscr, FALSE);
	endwin();
	resizeterm(groessen.ws_row,groessen.ws_col);
	starteCurses();
	wresize(haupt, groessen.ws_row, groessen.ws_col);
	tmp = berechneZeilenumbrueche();
	tmp2 = 81 / (groessen.ws_col - 2);
	if (3 == aktiv) {
		wresize(links, foh[aktiv] + tmp, (groessen.ws_col - 2) / 3);
		wresize(mitte, foh[aktiv] + tmp, (groessen.ws_col - 2) / 3);
		wresize(rechts, foh[aktiv] + tmp, (groessen.ws_col - 2) / 3);
	} else {
		wresize(links, foh[aktiv] + tmp, (groessen.ws_col - 0.5) / 2 - 1);
		wresize(rechts, foh[aktiv] + tmp, (groessen.ws_col - 0.5) / 2 - 1);
	}
	wresize(unten, groessen.ws_row - foh[aktiv] - 5 - tmp - tmp2, groessen.ws_col - 2);
	maleRahmen(getmaxy(haupt), getmaxx(haupt));
	fensterNeu = FALSE;
}

/* Verarbeitet die Antwort die nach einer Lueftereinstellung gesendet wird.
   Bei Erfolg wird True zurueckgegeben beim Scheitern False */
static bool verarbeiteAntwort() {
	int i = 0;
	int g = 0;
	int zaehler = 0;
	unsigned char buf[8] = {0};
	unsigned char buf2[8] = {0};
	const unsigned char ab[5] = {0x55,0x20,0x05,0x90,0x00};
	unsigned char versatz = 0;
	while (lauf) {
		if (datenDa) {
			datenDa = FALSE;
			/* Bytes in buf einlesen */
			sigprocmask(SIG_BLOCK,&mask,&orig_mask);
			g = (int)read(fd,buf,8);
			sigprocmask(SIG_SETMASK,&orig_mask,NULL);
			/* Anfangsbyte suchen */
			while (0x55 != buf[versatz] && versatz < g && 0 == zaehler) versatz++;
			if (versatz == g) {
				versatz = 0;
				if (8 == g) {
					continue;
				} else {
					break;
				}
			}
			/* Mal schauen was ankam */
			for (i = versatz; i < g; i++) {
				buf2[zaehler++] = buf[i];
			}
			if (4 < zaehler) {
				/* Antwort ueberpruefen */
				for (g = 0; g < 5; g++) {
					if (buf2[g] != ab[g]) {
						return FALSE;
					}
				}
				return TRUE;
			}
		}
		/* Auf Daten vom FOM pro warten */
		usleep(1000000);
		if (!datenDa) {
			/* Wenn keine Daten kamen abbrechen */
			break;
		}
	}
	return FALSE;
}

/* Tastenabfrage wenn im oberen Menue */
static bool frageTasteAb() {
	int eingabe = 0;
	bool rueck = FALSE;
	eingabe = wgetch(haupt);
	switch (eingabe) {
		case ERR:
			rueck = FALSE;
			break;
		case KEY_LEFT:
		case KEY_DOWN:
			menu = (5 + menu - 1) % 5;
			pthread_mutex_lock(&mutex);
			starteMenue();
			pthread_mutex_unlock(&mutex);
			rueck = FALSE;
			break;
		case KEY_RIGHT:
		case KEY_UP:
			menu = (menu + 1) % 5;
			pthread_mutex_lock(&mutex);
			starteMenue();
			pthread_mutex_unlock(&mutex);
			rueck = FALSE;
			break;
		case (int)'\n':
			pthread_mutex_lock(&mutex2);
			aktiv = menu;
			pthread_mutex_unlock(&mutex2);
			rueck = TRUE;
			break;
		case (int)'q':
			pthread_mutex_lock(&mutex2);
			lauf = FALSE;
			pthread_mutex_unlock(&mutex2);
			rueck = TRUE;
			break;
			/* Auskommentieren um zu schauen welche Codes welche Tasten haben */
			/*default:
				printw("Tasteneingabe: %d ", eingabe);*/
	}
	return rueck;
}

/*
static bool sendeAnfrage() {
	
}
*/

/* schaut ob das Programm beendet werden muss */
static bool syncP() {
	/* lauf synchronisieren. Es darf waehrenddessen aber
	 * auch nicht der Signalhandler dazwischenfunken */
	sigprocmask(SIG_BLOCK,&mask,&orig_mask);
	pthread_mutex_lock(&mutex2);
	if (!lauf) {
		pthread_mutex_unlock(&mutex2);
		sigprocmask(SIG_SETMASK,&orig_mask,NULL);
		return FALSE;
	}
	pthread_mutex_unlock(&mutex2);
	sigprocmask(SIG_SETMASK,&orig_mask,NULL);
	return TRUE;
}

/* schaut ob das Programm beendet werden muss und passt das Fenster an */
static bool syncMuP() {
	if (!syncP()) {
		return FALSE;
	}
	if (fensterNeu) {
		pthread_mutex_lock(&mutex);
		passeFensterAn();
		starteMenue();
		gebeFomMenueAus();
		gebeToolTippsAus();
		pthread_mutex_unlock(&mutex);
	}
	return TRUE;
}

/* schaut ob das Programm beendet werden muss und passt das Fenster an
 * und verwirft die Tasteneingabe Fenster neu */
static bool syncMuPuT() {
	if (!syncP()) {
		return FALSE;
	}
	if (fensterNeu) {
		pthread_mutex_lock(&mutex);
		passeFensterAn();
		starteMenue();
		gebeFomMenueAus();
		gebeToolTippsAus();
		getch();
		pthread_mutex_unlock(&mutex);
	}
	return TRUE;
}

/* Monitormodus: Zeigt Monitordaten des FOM pro an.
   Ausserdem koennen im manuellen Modus die Luefter
   ein- und ausgeschaltet werden.*/
static void frageMonitorDatenAb() {
	int i = 0;
	int g = 0;
	int zaehler = 0;
	short *zeiger = NULL;
	char buf[256] = {0};
	char buf2[256] = {0};
	const unsigned char ds[5] = {0x55,0x20,0x05,0x13,0x00};
	unsigned char versatz = 0;
	
	/* Programm beenden mit SIGINT oder SIGTERM */
	while (TRUE) {
		/* lauf synchronisieren. Es darf waehrenddessen aber
		 * auch nicht der Signalhandler dazwischenfunken */
		sigprocmask(SIG_BLOCK,&mask,&orig_mask);
		pthread_mutex_lock(&mutex2);
		if (!lauf || weiter) {
			pthread_mutex_unlock(&mutex2);
			sigprocmask(SIG_SETMASK,&orig_mask,NULL);
			break;
		}
		pthread_mutex_unlock(&mutex2);
		pthread_mutex_lock(&mutex3);
		if (-1 == write(fd,ds,5)) {
			gingNichtGut("write schlug fehl!");
		}
		sigprocmask(SIG_SETMASK,&orig_mask,NULL);
		while (syncMuPuT()) {
			if (datenDa) {
				datenDa = FALSE;
				/* Bytes lesen und ausgeben */
				sigprocmask(SIG_BLOCK,&mask,&orig_mask);
				g = (int)read(fd,buf,150);
				sigprocmask(SIG_SETMASK,&orig_mask,NULL);
				pthread_mutex_unlock(&mutex3);
				while (0x55 != buf[versatz] && versatz < g && 0 == zaehler) versatz++;
				if (versatz == g) {
					versatz = 0;
					if (150 == g) {
						pthread_mutex_lock(&mutex3);
						continue;
					} else {
						break;
					}
				}
				/* Mal schauen was ankam */
				for (i = versatz; i < g; i++) {
					buf2[zaehler++] = buf[i];
				}
				if (147 < zaehler) {
					/* Fomdaten aktuallisieren */
					for (i = 0; i < 16; i++) {
						/* Wenn gerade Einstellugen vorgenommen werden,
						 * werden diese nicht angepasst. */
						if (!(einstellungen && i ==
							luefterKanal[woBinIch / 10])) {
							lz[i] = (buf2[i * 4 + 7]) & 0x01;
							lspan[i] = buf2[i * 4 + 4];
						}
						lvb[i] = buf2[i * 4 + 5];
						lupm[i] = buf2[i * 4 + 6];
						zeiger = (short *)(buf2 + 68 + i * 2);
						stemp[i] = ((float)*zeiger) / 10;
					}
					zeiger = (short *)(buf2 + 110);
					durchfluss[0] = ((float)*zeiger) / 1000;
					zeiger = (short *)(buf2 + 112);
					durchfluss[1] = ((float)*zeiger) / 100;
					for (i = 0; i < 3; i++) {
						nts[i] = buf2[107 + i];
						datum[i] = buf2[100 + i];
					}
					for (i = 3; i < 7; i++) {
						datum[i] = buf2[100 + i];
					}
					/* Waehrend dieser Faden in die Ausgabe schreibt,
					* darf der andere dies nicht auch machen */
					pthread_mutex_lock(&mutex);
					gebeFomMenueAus();
					/* So dieser Faden ist mit der Ausgabe fertig,
					* der andere darf also auch wieder */
					pthread_mutex_unlock(&mutex);
					break;
				}
			}
			/* Eine Sekunde pennen der FOM schickt jede Sekunde
			   Das Gepenne wird durch SIGIO Signal unterbrochen */
			usleep(1000000);
			if (!datenDa) {
				pthread_mutex_unlock(&mutex3);
				break;
			}
		}
		zaehler = 0;
	}
}

/* Fom-Monitormodus
   Die Taste q beendet das Programm.
   Die Taste esc verwirft getaetigte Einstellungen
   Die Pfeiltasten navigieren durch das Menueeintraegen
   alle anderen Tasten werden ignoriert */
static void *starteMonitorModus(void *data) {
	int eingabe = 0;
	/* Bytefolge um Luefter an oder auszuschalten */
	unsigned char la[5] = {0x55,0x20,0x05,0x21,0x00};
	/* Bytefolge um die manuelle Spannung der Luefter zu setzen */
	unsigned char lms[6] = {0x55,0x20,0x06,0x20,0x00,0x00};
	/* Wird verwendet, um den Sprung von 3 Menueeintraegen
	 * auf 1 Menueeintrag zu speichern */
	unsigned char tmp = 0;
	/* Wird zum Speichern des Anfangsluefter bzw, -sensorkanals benutzt */
	unsigned char tmp2 = 0;
	/* Zwei weitere temporaere Variabeln */
	unsigned char tmp3 = 0;
	unsigned char tmp4 = 0;
	woBinIch = 0;

	while (syncMuP()) {
		do {
			pthread_mutex_lock(&mutex);
			gebeToolTippsAus();
			pthread_mutex_unlock(&mutex);
			tmp4 = woBinIch % 10;
			eingabe = mvgetch(0, 0);
			switch (eingabe) {
				case KEY_UP:
					if (45 == woBinIch) {
						woBinIch = 35 + tmp;
					} else if (woBinIch < 4) {
						tmp = woBinIch % 5;
						woBinIch = 60;
					} else if (4 == tmp4 ||
							60 == woBinIch) {
						woBinIch = (woBinIch + 80 - 10) % 80;
					} else {
						woBinIch = (woBinIch + 80 - 5) % 80;
					}
					break;
				case KEY_DOWN:
					if (4 == tmp4 || 50 == woBinIch) {
						woBinIch = (woBinIch + 10) % 80;
					} else if(60 == woBinIch) {
						woBinIch = tmp;
					} else if (35 == woBinIch ||
							36 == woBinIch ||
							37 == woBinIch) {
						tmp = woBinIch % 5;
						woBinIch = 45;
					} else {
						woBinIch = (woBinIch + 5) % 80;
					}
					break;
				case KEY_LEFT:
					if ((68 == woBinIch) || (78 == woBinIch)
							|| 58 == woBinIch) {
						continue;
					} else if (5 == tmp4) {
						woBinIch = (woBinIch + 5 - 2) % 5
								+ (woBinIch / 5) * 5;
					} else if (73 == woBinIch || 43 == woBinIch) {
						woBinIch = (woBinIch + 1) % 5
								+ (woBinIch / 5) * 5;
					} else if (48 == woBinIch || 53 == woBinIch
							|| 63 == woBinIch) {
						woBinIch = (woBinIch + 5 - 3) % 5
								+ (woBinIch / 5) * 5;
					} else {
						woBinIch = (woBinIch + 5 - 1) % 5
								+ (woBinIch / 5) * 5;
					}
					break;
				case KEY_RIGHT:
					if (68 == woBinIch || 78 == woBinIch
							|| 58 == woBinIch) {
						continue;
					} else if (8 == tmp4) {
						woBinIch = (woBinIch + 2) % 5 +
								(woBinIch / 5) * 5;
					} else if (74 == woBinIch || 44 == woBinIch) {
						woBinIch = (woBinIch + 5 - 1) % 5
								+ (woBinIch / 5) * 5;
					} else if (45 == woBinIch || 50 == woBinIch
							|| 60 == woBinIch) {
						woBinIch = (woBinIch + 3) % 5
								+ (woBinIch / 5) * 5;
					} else {
						woBinIch = (woBinIch + 1) % 5 +
								(woBinIch / 5) * 5;
					}
					break;
				case (int) '\n':
					tmp3 = woBinIch / 10;
					/* Tastenabfrage Luefter- oder Sensorkanalwechsel */
					if ((0 == tmp4) || (3 == tmp4)) {
						if (0 == tmp4) {
							tmp2 = luefterKanal[tmp3];
						} else if (3 == tmp4) {
							tmp2 = sensorKanal[tmp3];
						}
						while (syncMuP()) {
							eingabe = mvgetch(0, 0);
							pthread_mutex_lock(&mutex);
							switch (eingabe) {
								case KEY_UP:
								case KEY_RIGHT:
									if (0 == tmp4) {
										luefterKanal[tmp3] =
										(luefterKanal[tmp3] + 1) % 16;
									} else if (3 == tmp4) {
										sensorKanal[tmp3] =
										(sensorKanal[tmp3] + 1) % 16;
									}
									break;
								case KEY_DOWN:
								case KEY_LEFT:
									if (0 == tmp4) {
										luefterKanal[tmp3] =
										(luefterKanal[tmp3] + 16 - 1) % 16;
									} else if (3 == tmp4) {
										sensorKanal[tmp3] =
										(sensorKanal[tmp3] + 16 - 1) % 16;
									}
									break;
								case 27:
									if (0 == tmp4) {
										luefterKanal[tmp3] = tmp2;
									} else if (3 == tmp4) {
										sensorKanal[tmp3] = tmp2;
									}
									break;
							}
							gebeFomMenueAus();
							pthread_mutex_unlock(&mutex);
							if ((int)'\n' == eingabe || 27 == eingabe
								|| (int)'q' == eingabe) {
								break;
							}
						}
					} else if (1 == tmp4) {
						pthread_mutex_lock(&mutex);
						einstellungen = TRUE;
						pthread_mutex_unlock(&mutex);
						tmp3 = luefterKanal[tmp3];
						/* Wenn der Luefter nicht im manuellen Modus ist,
						 * kann man ihn auch nicht an oder ausschalten. */
						if (1 == lm[tmp3]) {
							break;
						}
						tmp2 = lz[tmp3];
						while (syncMuP()) {
							eingabe = mvgetch(0, 0);
							pthread_mutex_lock(&mutex);
							switch (eingabe) {
								case KEY_UP:
								case KEY_RIGHT:
								case KEY_DOWN:
								case KEY_LEFT:
									lz[tmp3] =
										(lz[tmp3] + 1) % 2;
									break;
								case 27:
									lz[tmp3] = tmp2;
									einstellungen = FALSE;
									break;
								/* TEST: Luefter ein- oder ausschalten */
								case (int)'\n':
									/* Luefter an oder aus */
									la[3] += lz[tmp3];
									/* Luefter auswaehlen */
									la[4] = tmp3;
									pthread_mutex_lock(&mutex3);
									do {
										sigprocmask(SIG_BLOCK,&mask,&orig_mask);
										if (-1 == write(fd,la,5)) {
											gingNichtGut("write schlug fehl!");
										}
										sigprocmask(SIG_SETMASK,&orig_mask,NULL);
									} while (!verarbeiteAntwort());
									einstellungen = FALSE;
									pthread_mutex_unlock(&mutex3);
							}
							gebeFomMenueAus();
							pthread_mutex_unlock(&mutex);
							if (27 == eingabe || (int)'\n' == eingabe
								|| (int)'q' == eingabe) {
								break;
							}
						}
					/* TEST: Luefterspannung einstellen */
					} else if (5 == tmp4) {
						/* Wenn der Luefter nicht im manuellen Modus ist,
						 * kann man die Spannung nicht veraendern. */
						if (1 == lm[tmp3]) {
							break;
						}
						pthread_mutex_lock(&mutex);
						einstellungen = TRUE;
						pthread_mutex_unlock(&mutex);
						tmp3 = luefterKanal[tmp3];
						tmp2 = lspan[tmp3];
						while (syncMuP()) {
							eingabe = mvgetch(0, 0);
							pthread_mutex_lock(&mutex);
							switch (eingabe) {
								case KEY_UP:
								case KEY_RIGHT:
									lspan[tmp3] =
										(lspan[tmp3] + 1) % 121;
									break;
								case KEY_DOWN:
								case KEY_LEFT:
									lspan[tmp3] =
										(lspan[tmp3] + 121 - 1) % 121;
									break;
								case 27:
									lspan[tmp3] = tmp2;
									einstellungen = FALSE;
									break;
								/* TEST: manuelle Luefterspannung einstellen */
								case (int)'\n':
									/* Luefter auswaehlens */
									lms[4] = tmp3;
									/* Spannung setzen */
									lms[5] = lspan[tmp3];
									pthread_mutex_lock(&mutex3);
									do {
										sigprocmask(SIG_BLOCK,&mask,&orig_mask);
										if (-1 == write(fd,lms,6)) {
											gingNichtGut("write schlug fehl!");
										}
										sigprocmask(SIG_SETMASK,&orig_mask,NULL);
									} while (!verarbeiteAntwort());
									einstellungen = FALSE;
									pthread_mutex_unlock(&mutex3);
							}
							gebeFomMenueAus();
							pthread_mutex_unlock(&mutex);
							if (27 == eingabe || (int)'\n' == eingabe
								|| (int)'q' == eingabe) {
								break;
							}
						}
					}
					/* Wenn die Taste q gedrueckt wird, wird das Programm beendet
					 * andernfalls wird die Eingabeschleife fortgesetzt */
					if ((int)'q' != eingabe) {
						continue;
					}
				case (int)'q':
					sigprocmask(SIG_BLOCK,&mask,&orig_mask);
					pthread_mutex_lock(&mutex2);
					lauf = FALSE;
					pthread_mutex_unlock(&mutex2);
					sigprocmask(SIG_SETMASK,&orig_mask,NULL);
					return data;
				case 27:
					while(!frageTasteAb());
					pthread_mutex_lock(&mutex2);
					weiter = TRUE;
					pthread_mutex_unlock(&mutex2);
					return data;
			}
			pthread_mutex_lock(&mutex);
			gebeFomMenueAus();
			pthread_mutex_unlock(&mutex);
		} while (FALSE);
	}
	return data;
}

/* sendet an den FOMpro die Monitorabfrage und speichert das Ergebnis */
static void *frageFOMDatenAb(void *data) {
	int i = 0;
	int g = 0;
	int zaehler = 0;
	short *zeiger = NULL;
	char buf[256] = {0};
	char buf2[256] = {0};
	const unsigned char ds[5] = {0x55,0x20,0x05,0x13,0x00};
	unsigned char versatz = 0;
	
	/* Programm beenden mit SIGINT oder SIGTERM */
	while (TRUE) {
		/* lauf synchronisieren. Es darf waehrenddessen aber
		 * auch nicht der Signalhandler dazwischenfunken */
		sigprocmask(SIG_BLOCK,&mask,&orig_mask);
		pthread_mutex_lock(&mutex2);
		if (!lauf || weiter) {
			pthread_mutex_unlock(&mutex2);
			sigprocmask(SIG_SETMASK,&orig_mask,NULL);
			break;
		}
		pthread_mutex_unlock(&mutex2);
		pthread_mutex_lock(&mutex3);
		if (-1 == write(fd,ds,5)) {
			pthread_mutex_unlock(&mutex3);
		}
		sigprocmask(SIG_SETMASK,&orig_mask,NULL);
		while (syncP()) {
			if (datenDa) {
				datenDa = FALSE;
				sigprocmask(SIG_BLOCK,&mask,&orig_mask);
				/* Bytes lesen und ausgeben */
				g = (int)read(fd,buf,150);
				sigprocmask(SIG_SETMASK,&orig_mask,NULL);
				while (0x55 != buf[versatz] && versatz < g && 0 == zaehler) versatz++;
				if (versatz == g) {
					versatz = 0;
					if (150 == g) {
						continue;
					} else {
						break;
					}
				}
				/* Mal schauen was ankam */
				for (i = versatz; i < g; i++) {
					buf2[zaehler++] = buf[i];
				}
				if (147 < zaehler) {
					/* Fomdaten aktuallisieren */
					for (i = 0; i < 16; i++) {
						lz[i] = (buf2[i * 4 + 7]) & 0x01;
						pthread_mutex_lock(&mutex);
						if (!einstellungen) {
							lspan[i] = buf2[i * 4 + 4];
						}
						pthread_mutex_unlock(&mutex);
						lvb[i] = buf2[i * 4 + 5];
						lupm[i] = buf2[i * 4 + 6];
						zeiger = (short *)(buf2 + 68 + i * 2);
						stemp[i] = ((float)*zeiger) / 10;
					}
					zeiger = (short *)(buf2 + 110);
					durchfluss[0] = ((float)*zeiger) / 1000;
					zeiger = (short *)(buf2 + 112);
					durchfluss[1] = ((float)*zeiger) / 100;
					for (i = 0; i < 3; i++) {
						nts[i] = buf2[107 + i];
						datum[i] = buf2[100 + i];
					}
					for (i = 3; i < 7; i++) {
						datum[i] = buf2[100 + i];
					}
					break;
				}
			}
			/* Eine Sekunde pennen der FOM schickt jede Sekunde
			   Das Gepenne wird durch SIGIO Signal unterbrochen */
			usleep(1000000);
			if (!datenDa) {
				break;
			}
		}
		pthread_mutex_unlock(&mutex3);
		zaehler = 0;
		/* Eine kurze Pause einlegen um dem anderen Faden
		 * zu ermoeglichen mit dem FOM zu kommunizieren */
		usleep(100);
	}
	return data;
}

/* schaut ob das Programm beendet werden muss und passt das Fenster an */
static bool syncFuP(unsigned char lk) {
	if (!syncP()) {
		return FALSE;
	}
	if (fensterNeu) {
		passeFensterAn();
		starteMenue();
		maleLuefterKonfMenue(lk);
		gebeToolLKTippsAus();
	}
	return TRUE;
}

/* Speichert eine Lueftereinstellung dauerhaft.
   Bei Erfolg wird True zurueckgegeben beim Scheitern False */
static bool speichereLE() {
	const unsigned char se[5] = {0x55,0x20,0x05,0x92,0x00};
	/* Auf Antwort vom FOMpro warten */
	if(!verarbeiteAntwort()) {
		return FALSE;
	}
	do {
		/* sende Einstellungen speichern Bytefolge */
		sigprocmask(SIG_BLOCK,&mask,&orig_mask);
		if (-1 == write(fd,se,5)) {
			gingNichtGut("write schlug fehl!");
		}
		sigprocmask(SIG_SETMASK,&orig_mask,NULL);
		
	/* wieder auf Antwort vom FOMpro warten,
	 * wenn es schief ging erneut anfragen */
	} while(!verarbeiteAntwort());
	return TRUE;
}

/* Hilfsfunktion fuer Einstellungen */
static void helfeglkma(unsigned char zble[16], unsigned char kanal, unsigned char alarmKonf[7],
		unsigned char minWert, unsigned char maxWert) {
	int eingabe = 0;
	unsigned char sicherung = zble[kanal];
	while (syncFuP(kanal)) {
		eingabe = mvgetch(0, 0);
		switch (eingabe) {
			case KEY_UP:
			case KEY_RIGHT:
				zble[kanal] = (zble[kanal] - minWert +  1) % maxWert + minWert;
				break;
			case KEY_DOWN:
			case KEY_LEFT:
				zble[kanal] = (zble[kanal] - minWert +  maxWert - 1) % maxWert + minWert;
				break;
			case 27:
				zble[kanal] = sicherung;
				break;
		}
		if (1 == aktiv) {
			maleLuefterKonfMenue(kanal);
		} else if (2 == aktiv) {
			if (kanal < 16) {
				maleSensorKonfMenue(kanal);
			} else {
				maleSensorKonfMenue(kanal - 16);
			}
		} else if (3 == aktiv) {
			maleReglerKonfMenue(kanal);
		}
		if ('\n' == eingabe) {
			mvwaddnstr(unten, 0, COLS / 2 - 17, "Einstellung wird gespeichert ...", 33);
			wrefresh(unten);
			alarmKonf[6] = zble[kanal];
			pthread_mutex_lock(&mutex3);
			do {
				sigprocmask(SIG_BLOCK,&mask,&orig_mask);
				if (-1 == write(fd,alarmKonf,7)) {
					gingNichtGut("write schlug fehl!");
				}
				sigprocmask(SIG_SETMASK,&orig_mask,NULL);
			} while (!speichereLE());
			mvwaddnstr(unten, 0, COLS / 2 - 17, " Einstellung wurde gespeichert. ", 33);
			wrefresh(unten);
			pthread_mutex_unlock(&mutex3);
			return;
		} else if (27 == eingabe) {
			return;
		}
	}
}

/*Reiter fuer die Konfiguration der Luefter */
static void gebeLuefterKonfMenueAus() {
	int eingabe = 0;
	int eingabe2 = 0;
	int zu = 0;
	int zeile = 0;
	int spalte = 0;
	int zaehler = 0;
	char buf[256] = {0};
	char buf2[256] = {0};
	/* Bytefolge fuer die Namensvergabe der Luefter */
	unsigned char sendeNamen[26] = {0x55,0x20,0x1A,0x50,0x01,0x5E,
		0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
		0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20};
	/* Bytefolge um Lueftereinstellungen zu setzen */
	unsigned char alarmKonf[7] = {0x55,0x20,0x07,0x90,0x00,0x02,0x00};
	/* Bytefolge um die manuelle Spannung der Luefter zu setzen */
	unsigned char lms[6] = {0x55,0x20,0x06,0x20,0x00,0x00};
	char name[21] = {0};
	unsigned char i = 0;
	unsigned char woWarIch = 0;
	unsigned char lk = 0; /* Luefterkanal */
	unsigned char sicherung = 0;
	woBinIch = 0;

	while(syncFuP(lk)) {
		eingabe = mvgetch(0, 0);
		switch(eingabe) {
			case KEY_RIGHT:
				if (0 == woBinIch) {
					woBinIch = 1;
				} else if (18 == woBinIch) {
					continue;
				} else if (0 == woBinIch % 3) {
					woBinIch = (woBinIch + 2) % 3
						+ (woBinIch / 3) * 3;
				} else {
					woBinIch = (woBinIch + 1) % 3
						+ (woBinIch / 3) * 3;
				}
				maleLuefterKonfMenue(lk);
				gebeToolLKTippsAus();
				break;
			case KEY_LEFT:
				if (2 == woBinIch) {
					woBinIch = 1;
				} else if (2 == woBinIch % 3) {
					woBinIch = (woBinIch + 3 - 2) % 3
						+ (woBinIch / 3) * 3;
				} else if (18 == woBinIch) {
					continue;
				} else {
					woBinIch = (woBinIch + 3 - 1) % 3
						+ (woBinIch / 3) * 3;
				}
				maleLuefterKonfMenue(lk);
				gebeToolLKTippsAus();
				break;
			case KEY_UP:
				if (0 == woBinIch || 1 == woBinIch) {
					woWarIch = woBinIch;
					woBinIch = 18;
				} else if (3 == woBinIch) {
					woBinIch = woWarIch;
				} else if (2 == woBinIch) {
					woBinIch = 17;
				} else {
					woBinIch = (woBinIch + 21 - 3) % 21;
				}
				maleLuefterKonfMenue(lk);
				gebeToolLKTippsAus();
				break;
			case KEY_DOWN:
				if (0 == woBinIch || 1 == woBinIch) {
					woWarIch = woBinIch;
					woBinIch = 3;
				} else if (17 == woBinIch) {
					woBinIch = 2;
				} else if (18 == woBinIch) {
					woBinIch = woWarIch;
				} else {
					woBinIch = (woBinIch + 3) % 21;
				}
				maleLuefterKonfMenue(lk);
				gebeToolLKTippsAus();
				break;
			/* ESC Taste ab ins obere Menue */
			case 27:
				while(!frageTasteAb());
				pthread_mutex_lock(&mutex2);
				weiter = TRUE;
				pthread_mutex_unlock(&mutex2);
				return;
			case (int) '\n':
				switch (woBinIch) {
					case 0:
						/* Luefterkanalwechsel */
						sicherung = lk;
						while(syncFuP(lk)) {
							eingabe = mvgetch(0, 0);
							switch (eingabe) {
								case KEY_UP:
								case KEY_RIGHT:
									lk = (lk + 1) % 16;
									break;
								case KEY_DOWN:
								case KEY_LEFT:
									lk = (lk + 16 - 1) % 16;
									break;
								case 27:
									lk = sicherung;
									break;
							}
							maleLuefterKonfMenue(lk);
							if ('\n' == eingabe || 27 == eingabe) {
								break;
							}
						}
						break;
					case 1:
						/* aktuellen Namen sichern */
						memcpy(name, namen[lk], 20);
						/* Namen fuer die Eingabe loeschen */
						memset(namen[lk], ' ', 20);
						maleLuefterKonfMenue(lk);
						wattron(links, COLOR_PAIR(2));
						curs_set(2);
						zu = COLS / 2 - 2;
						for (i = 0; i < 21; i++) {
							zeile = (9 + i) / zu;
							spalte = (9 + i) % zu;
							eingabe = mvwgetch(links, zeile, spalte);
							switch (eingabe) {
								case KEY_UP:
								case KEY_DOWN:
								case KEY_LEFT:
								case KEY_RIGHT:
									i--;
									break;
								case 194:
								case 195:
									eingabe2 = wgetch(links);
									namen[lk][i] = ausZweiMachEins(eingabe,
											eingabe2, zeile, spalte, links);

									break;
								case KEY_BACKSPACE:
								case 127:
									if (0 == i) {
										i--;
									} else {
										i = i - 2;
									}
									mvwaddch(links, zeile, spalte - 1, ' ');
									namen[lk][i - 1] = ' ';
									break;
								case 27:
									break;
								case KEY_RESIZE:
								case ERR:
									sigprocmask(SIG_BLOCK,&mask,&orig_mask);
									if (fensterNeu) {
										passeFensterAn();
										starteMenue();
										for (eingabe2 = (int) i; eingabe2 < 20; eingabe2++) {
											namen[lk][eingabe2] = ' ';
										}
										namen[lk][20] = '\0';
										maleLuefterKonfMenue(lk);
										gebeToolLKTippsAus();
										wattron(links, COLOR_PAIR(2));
										curs_set(2);
									}
									sigprocmask(SIG_SETMASK,&orig_mask,NULL);
									i--;
									continue;
								default:
									if ((int)'\n' == eingabe) {
										for (eingabe2 = (int) i; eingabe2 < 20; eingabe2++) {
											namen[lk][eingabe2] = ' ';
										}
										namen[lk][20] = '\n';
										break;
									}
									namen[lk][i] = (char) eingabe;
									if (20 == i) {
										mvwaddstr(unten, 0, COLS / 2 - 14,
												"21. Zeichen war kein enter!");
										wrefresh(unten);
										break;
									}
									mvwaddch(links, zeile, spalte, eingabe);
							}
							if (27 == eingabe || (int) '\n' == eingabe) {
								break;
							}
							wrefresh(links);
						}
						if (27 == eingabe) {
							strncpy(namen[lk], name, 20);
							namen[lk][20] = '\0';
						} else if ('\n' == namen[lk][20]) {
							namen[lk][20] = '\0';
							sendeNamen[5] = 0x5E + lk;
							strncpy((char *)&sendeNamen[6], namen[lk], 20);
							pthread_mutex_lock(&mutex3);
							sigprocmask(SIG_BLOCK,&mask,&orig_mask);
							if (-1 == write(fd, sendeNamen, 26)) {
								pthread_mutex_unlock(&mutex3);
								gingNichtGut("write schlug fehl!");
							}
							sigprocmask(SIG_SETMASK,&orig_mask,NULL);
							while (syncP()) {
								if (datenDa) {
									datenDa = FALSE;
									sigprocmask(SIG_BLOCK,&mask,&orig_mask);
									eingabe = read(fd, buf, 8);
									sigprocmask(SIG_SETMASK,&orig_mask,NULL);
									for (i = 0; i < eingabe; i++) {
										buf2[zaehler++] = buf[i];
										fprintf(stderr, "%x ", buf2[zaehler - 1]);
									}
									if (5 < zaehler) {
										mvwprintw(unten, 0, COLS / 2 - 10,
											"Lüftername %d ", lk + 1);
										if (0x55 == buf2[0] && 0x20 == buf2[1] &&
											0x06 == buf2[2] && 0x50 == buf2[3] &&
											0x01 == buf2[4] && (0x5E + lk) ==
											(unsigned char)buf2[5]) {
											waddstr(unten, "wurde übernommen.");
										} else {
											waddstr(unten,
												"konnte nicht übernommen werden.");
										}
										wrefresh(unten);
										zaehler = 0;
										break;
									}
								}
								usleep(250000);
								if (!datenDa) {
									pthread_mutex_unlock(&mutex3);
									gingNichtGut("Verbindung zum FOM wurde gestört!");
								}
							}
							pthread_mutex_unlock(&mutex3);
						}
						wattroff(links, COLOR_PAIR(2));
						curs_set(0);
						maleLuefterKonfMenue(lk);
						break;
					case 3:
						/* Lueftermoduswechsel */
						alarmKonf[5] = lk * 14 + 1;
						helfeglkma(lm,lk, alarmKonf, 0, 2);
						break;
					case 5:
						/* Impulse pro Umdrehung Einstellung */
						alarmKonf[5] = lk * 14 + 7;
						helfeglkma(ipu,lk, alarmKonf, 2, 31);
						break;
					case 6:
						/* Alarm Meldung Einstellung */
						alarmKonf[5] = lk * 14 + 2;
						helfeglkma(ameld,lk, alarmKonf, 0, 100);
						break;
					case 8:
						/* Minimalspannung einstellen */
						alarmKonf[5] = lk * 14 + 9;
						helfeglkma(minu,lk, alarmKonf, 0, 121);
						break;
					case 9:
						/* Alarm Mainboard Einstellung */
						alarmKonf[5] = lk * 14 + 3;
						helfeglkma(amobo,lk, alarmKonf, 0, 100);
						break;
					case 11:
						/* Maximalspannung einstellen */
						alarmKonf[5] = lk * 14 + 10;
						helfeglkma(maxu,lk, alarmKonf, 0, 121);
						break;
					case 12:
						/* Alarm Herunterfahren Einstellung */
						alarmKonf[5] = lk * 14 + 4;
						helfeglkma(areli1,lk, alarmKonf, 0, 100);
						break;
					case 14:
						/* Anlaufspannung einstellen */
						alarmKonf[5] = lk * 14 + 8;
						helfeglkma(boost,lk, alarmKonf, 0, 121);
						break;
					case 15:
						/* Alarm Notabschaltung Einstellung */
						alarmKonf[5] = lk * 14 + 5;
						helfeglkma(areli2,lk, alarmKonf, 0, 100);
						break;
					case 17:
						/* Luefterspannung einstellen */
						/* Wenn der Luefter nicht im manuellen Modus ist
						 * kann die Spannung nicht eingestellt werden */
						if (1 == lm[lk]) {
							break;
						}
						sicherung = lspan[lk];
						pthread_mutex_lock(&mutex);
						einstellungen = TRUE;
						pthread_mutex_unlock(&mutex);
						while (syncFuP(lk)) {
							eingabe = mvgetch(0, 0);
							switch (eingabe) {
								case KEY_UP:
								case KEY_RIGHT:
									if (0 == lspan[lk]) {
										lspan[lk] = minu[lk] == 0 ? 1 : minu[lk];
									} else {
										lspan[lk] =
										(lspan[lk] - minu[lk] + 1) %
										(maxu[lk] - minu[lk] + 1) + minu[lk];
									}
									break;
								case KEY_DOWN:
								case KEY_LEFT:
									if (0 == lspan[lk]) {
										lspan[lk] = maxu[lk];
									} else {
										i = maxu[lk] - minu[lk] + 1;
										lspan[lk] =
										(lspan[lk] - minu[lk] + i - 1) % i + minu[lk];
									}
									break;
								case 27:
									lspan[lk] = sicherung;
									break;
							}
							maleLuefterKonfMenue(lk);
							if ('\n' == eingabe) {
								/* Luefter auswaehlens */
								lms[4] = lk;
								/* Spannung setzen */
								lms[5] = lspan[lk];
								pthread_mutex_lock(&mutex3);
								/* TEST: manuelle Luefterspannung einstellen */
								do {
									sigprocmask(SIG_BLOCK,&mask,&orig_mask);
									if (-1 == write(fd,lms,6)) {
										gingNichtGut("write schlug fehl!");
									}
									sigprocmask(SIG_SETMASK,&orig_mask,NULL);
								} while (!verarbeiteAntwort());
								pthread_mutex_unlock(&mutex3);
								break;
							} else if (27 == eingabe) {
								break;
							}
						}
						pthread_mutex_lock(&mutex);
						einstellungen = FALSE;
						pthread_mutex_unlock(&mutex);
						break;
					case 18:
						/* Alarm 12V Switch Einstellung */
						alarmKonf[5] = lk * 14 + 6;
						helfeglkma(a12V,lk, alarmKonf, 0, 100);
						break;
				}
				break;
			/* q Taste: Programm beenden */
			case (int)'q':
				sigprocmask(SIG_BLOCK,&mask,&orig_mask);
				pthread_mutex_lock(&mutex2);
				lauf = FALSE;
				pthread_mutex_unlock(&mutex2);
				sigprocmask(SIG_SETMASK,&orig_mask,NULL);
				return;
			
		}
	}
}

/* schaut ob das Programm beendet werden muss und passt das Fenster an */
static bool syncFuS(unsigned char sk) {
	if (!syncP()) {
		return FALSE;
	}
	if (fensterNeu) {
		passeFensterAn();
		starteMenue();
		maleSensorKonfMenue(sk);
		gebeToolSKTippsAus();
	}
	return TRUE;
}

/*Reiter fuer die Konfiguration der Sensoren */
static void gebeSensorKonfMenueAus() {
	int eingabe = 0;
	int eingabe2 = 0;
	int zu = 0;
	int zeile = 0;
	int spalte = 0;
	int zaehler = 0;
	short *zeiger;
	char buf[256] = {0};
	char buf2[256] = {0};
	/* Bytefolge fuer die Namensvergabe der Sensoren */
	unsigned char sendeNamen[26] = {0x55,0x20,0x1A,0x50,0x01,0x6E,
		0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
		0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20};
	/* Bytefolge um Sensoreinstellungen zu setzen */
	unsigned char alarmKonf[7] = {0x55,0x20,0x07,0x90,0x00,0xE1,0x00};
	char name[21] = {0};
	unsigned char i = 0;
	unsigned char sk = 0; /* Sensorkanal */
	unsigned char sicherung = 0;
	unsigned char woWarIch = 0;
	woBinIch = 0;
	zeiger = (short *)&alarmKonf[5];
	
	/* Sensorauswahl, Name, Device, 5 Alarme, Korrektur */
	while(syncFuS(sk)) {
		eingabe = mvgetch(0, 0);
		switch(eingabe) {
			case KEY_RIGHT:
				if (0 == woBinIch) {
					woBinIch = (woBinIch + 1) % 3 +
						(woBinIch / 3) * 3;
					woWarIch = woBinIch;
				} else if (0 == woBinIch % 3) {
					woBinIch = (woBinIch + 2) % 3 +
						(woBinIch / 3) * 3;
				} else {
					woBinIch = (woBinIch + 1) % 3 +
						(woBinIch / 3) * 3;
				}
				maleSensorKonfMenue(sk);
				gebeToolSKTippsAus();
				break;
			case KEY_LEFT:
				if (1 == woBinIch) {
					woBinIch = (woBinIch + 3 - 1) % 3 +
						(woBinIch / 3) * 3;
					woWarIch = woBinIch;
				} else if (2 == woBinIch || 0 == woBinIch % 3) {
					woBinIch = (woBinIch + 3 - 1) % 3 +
						(woBinIch / 3) * 3;
				} else {
					woBinIch = (woBinIch + 3 - 2) % 3 +
						(woBinIch / 3) * 3;
				}
				maleSensorKonfMenue(sk);
				gebeToolSKTippsAus();
				break;
			case KEY_UP:
				if (1 == woBinIch) {
					woBinIch = 9;
				} else if (3 == woBinIch) {
					woBinIch = woWarIch;
				} else {
					woBinIch = (woBinIch + 12 - 3) % 12;
				}
				maleSensorKonfMenue(sk);
				gebeToolSKTippsAus();
				break;
			case KEY_DOWN:
				if (1 == woBinIch) {
					woBinIch = 3;
				} else if (9 == woBinIch) {
					woBinIch = woWarIch;
				} else {
					woBinIch = (woBinIch + 3) % 12;
				}
				maleSensorKonfMenue(sk);
				gebeToolSKTippsAus();
				break;
			case (int) '\n':
				switch (woBinIch) {
					case 0:
						/* Sensorkanalwechsel */
						sicherung = sk;
						while(syncFuS(sk)) {
							eingabe = mvgetch(0, 0);
							switch (eingabe) {
								case KEY_UP:
								case KEY_RIGHT:
									sk = (sk + 1) % 16;
									break;
								case KEY_DOWN:
								case KEY_LEFT:
									sk = (sk + 16 - 1) % 16;
									break;
								case 27:
									sk = sicherung;
									break;
							}
							maleSensorKonfMenue(sk);
							if ('\n' == eingabe || 27 == eingabe) {
								break;
							}
						}
						break;
					case 1:
						/* aktuellen Namen sichern */
						memcpy(name, namen[sk + 16], 20);
						/* Namen fuer die Eingabe loeschen */
						memset(namen[sk + 16], ' ', 20);
						maleSensorKonfMenue(sk);
						wattron(links, COLOR_PAIR(2));
						curs_set(2);
						zu = COLS / 2 - 2;
						for (i = 0; i < 21; i++) {
							zeile = (9 + i) / zu;
							spalte = (9 + i) % zu;
							eingabe = mvwgetch(links, zeile, spalte);
							switch (eingabe) {
								case KEY_UP:
								case KEY_DOWN:
								case KEY_LEFT:
								case KEY_RIGHT:
									i--;
									break;
								case 194:
								case 195:
									eingabe2 = wgetch(links);
									namen[sk + 16][i] = ausZweiMachEins(eingabe,
											eingabe2, zeile, spalte, links);

									break;
								case KEY_BACKSPACE:
								case 127:
									if (0 == i) {
										i--;
									} else {
										i = i - 2;
									}
									mvwaddch(links, zeile, spalte - 1, ' ');
									namen[sk + 16][i - 1] = ' ';
									break;
								case 27:
									break;
								case KEY_RESIZE:
								case ERR:
									sigprocmask(SIG_BLOCK,&mask,&orig_mask);
									if (fensterNeu) {
										passeFensterAn();
										starteMenue();
										for (eingabe2 = (int) i; eingabe2 < 20; eingabe2++) {
											namen[sk + 16][eingabe2] = ' ';
										}
										namen[sk + 16][20] = '\0';
										maleSensorKonfMenue(sk);
										gebeToolSKTippsAus();
										wattron(links, COLOR_PAIR(2));
										curs_set(2);
									}
									sigprocmask(SIG_SETMASK,&orig_mask,NULL);
									i--;
									continue;
								default:
									if ((int)'\n' == eingabe) {
										for (eingabe2 = (int) i; eingabe2 < 20; eingabe2++) {
											namen[sk + 16][eingabe2] = ' ';
										}
										namen[sk + 16][20] = '\n';
										break;
									}
									namen[sk + 16][i] = (char) eingabe;
									if (20 == i) {
										mvwaddstr(unten, 0, COLS / 2 - 14,
												"21. Zeichen war kein enter!");
										wrefresh(unten);
										break;
									}
									mvwaddch(links, zeile, spalte, eingabe);
							}
							if (27 == eingabe || (int) '\n' == eingabe) {
								break;
							}
							wrefresh(links);
						}
						if (27 == eingabe) {
							strncpy(namen[sk + 16], name, 20);
							namen[sk + 16][20] = '\0';
						} else if ('\n' == namen[sk + 16][20]) {
							namen[sk + 16][20] = '\0';
							sendeNamen[5] = 0x6E + sk;
							strncpy((char *)&sendeNamen[6], namen[sk + 16], 20);
							pthread_mutex_lock(&mutex3);
							sigprocmask(SIG_BLOCK,&mask,&orig_mask);
							if (-1 == write(fd, sendeNamen, 26)) {
								pthread_mutex_unlock(&mutex3);
								gingNichtGut("write schlug fehl!");
							}
							sigprocmask(SIG_SETMASK,&orig_mask,NULL);
							while (syncP()) {
								if (datenDa) {
									datenDa = FALSE;
									sigprocmask(SIG_BLOCK,&mask,&orig_mask);
									eingabe = read(fd, buf, 8);
									sigprocmask(SIG_SETMASK,&orig_mask,NULL);
									for (i = 0; i < eingabe; i++) {
										buf2[zaehler++] = buf[i];
										fprintf(stderr, "%x ", buf2[zaehler - 1]);
									}
									if (5 < zaehler) {
										mvwprintw(unten, 0, COLS / 2 - 10,
											"Sensorname %d ", sk + 1);
										if (0x55 == buf2[0] && 0x20 == buf2[1] &&
											0x06 == buf2[2] && 0x50 == buf2[3] &&
											0x01 == buf2[4] && (0x6E + sk) ==
											(unsigned char)buf2[5]) {
											waddstr(unten, "wurde übernommen.");
										} else {
											waddstr(unten,
												"konnte nicht übernommen werden.");
										}
										wrefresh(unten);
										zaehler = 0;
										break;
									}
								}
								usleep(250000);
								if (!datenDa) {
									pthread_mutex_unlock(&mutex3);
									gingNichtGut("Verbindung zum FOM wurde gestört!");
								}
							}
							pthread_mutex_unlock(&mutex3);
						}
						wattroff(links, COLOR_PAIR(2));
						curs_set(0);
						maleSensorKonfMenue(sk);
						break;
					case 3:
						/* Alarm Meldung Einstellung */
						*zeiger = sk * 10 + 0x00E1;
						helfeglkma(ameld,sk + 16, alarmKonf, 0, 100);
						break;
					case 5:
						/* Alarm Mainboard Einstellung */
						*zeiger = sk * 10 + 0x00E2;
						helfeglkma(amobo,sk + 16, alarmKonf, 0, 100);
						break;
					case 6:
						/* Alarm Herunterfahren Einstellung */
						*zeiger = sk * 10 + 0x00E3;
						helfeglkma(areli1,sk + 16, alarmKonf, 0, 100);
						break;
					case 8:
						/* Alarm Notabschaltung Einstellung */
						*zeiger = sk * 10 + 0x00E4;
						helfeglkma(areli2,sk + 16, alarmKonf, 0, 100);
						break;
					case 9:
						/* Alarm 12V Switch Einstellung */
						*zeiger = sk * 10 + 0x00E5;
						helfeglkma(a12V,sk + 16, alarmKonf, 0, 100);
						break;
					case 11:
						/* Sensorkorrektureinstellung */
						*zeiger = sk * 10 + 0x00E6;
						helfeglkma(ks,sk, alarmKonf, 0, 241);
						break;
				}
				break;
			/* ESC Taste ab ins obere Menue */
			case 27:
				while(!frageTasteAb());
				pthread_mutex_lock(&mutex2);
				weiter = TRUE;
				pthread_mutex_unlock(&mutex2);
				return;
			/* q Taste: Programm beenden */
			case (int)'q':
				sigprocmask(SIG_BLOCK,&mask,&orig_mask);
				pthread_mutex_lock(&mutex2);
				lauf = FALSE;
				pthread_mutex_unlock(&mutex2);
				sigprocmask(SIG_SETMASK,&orig_mask,NULL);
				return;
		}
	}
}

/* schaut ob das Programm beendet werden muss und passt das Fenster an */
static bool syncFuR(unsigned char rk) {
	if (!syncP()) {
		return FALSE;
	}
	if (fensterNeu) {
		passeFensterAn();
		starteMenue();
		maleReglerKonfMenue(rk);
		gebeToolRTippsAus();
	}
	return TRUE;
}

/*Reiter fuer die Konfiguration der Regler */
static void gebeReglerKonfMenueAus() {
	int eingabe = 0;
	int eingabe2 = 0;
	int zu = 0;
	int zeile = 0;
	int spalte = 0;
	int zaehler = 0;
	short *zeiger;
	char buf[256] = {0};
	char buf2[256] = {0};
	/* Bytefolge fuer die Namensvergabe der Regulatoren */
	unsigned char sendeNamen[26] = {0x55,0x20,0x1A,0x50,0x01,0x7E,
		0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
		0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20};
	/* Bytefolge um Regulatoreinstellungen zu setzen */
	unsigned char reglerKonf[7] = {0x55,0x20,0x07,0x90,0x01,0x82,0x00};
	char name[21] = {0};
	unsigned char i = 0;
	unsigned char rk = 0; /* Regulatorkanal */
	unsigned char sicherung = 0;
	unsigned char woWarIch = 0;
	woBinIch = 0;
	zeiger = (short *)&reglerKonf[5];
	
	while(syncFuR(rk)) {
		eingabe = mvgetch(0, 0);
		switch(eingabe) {
			case KEY_RIGHT:
				if (0 == woBinIch) {
					woBinIch = 1;
					woWarIch = 1;
				} else if (3 == woBinIch) {
					woBinIch = 0;
					woWarIch = 0;
				} else if (0 == woBinIch % 5) {
					woBinIch = (woBinIch + 2) % 5 +
						(woBinIch / 5) * 5;
				} else if (22 == woBinIch) {
					woBinIch = 20;
				} else if (27 == woBinIch) {
					continue;
				} else {
					woBinIch = (woBinIch + 1) % 5 +
						(woBinIch / 5) * 5;
				}
				maleReglerKonfMenue(rk);
				gebeToolRTippsAus();
				break;
			case KEY_LEFT:
				if (0 == woBinIch) {
					woBinIch = 3;
				} else if (1 == woBinIch) {
					woBinIch = 0;
					woWarIch = 0;
				} else if (2 == woBinIch) {
					woBinIch = 1;
					woWarIch = 1;
				} else if (20 == woBinIch) {
					woBinIch = 22;
				} else if (27 == woBinIch) {
					continue;
				} else if (2 == woBinIch % 5) {
					woBinIch = (woBinIch + 5 - 2) % 5 +
						(woBinIch / 5) * 5;
				} else {
					woBinIch = (woBinIch + 5 - 1) % 5 +
						(woBinIch / 5) * 5;
				}
				maleReglerKonfMenue(rk);
				gebeToolRTippsAus();
				if (0 == woBinIch) {
					woWarIch = woBinIch;
				}
				break;
			case KEY_UP:
				if (0 == woBinIch || 1 == woBinIch) {
					woBinIch = 20;
				} else if (3 == woBinIch) {
					woBinIch = 18;
				} else if (5 == woBinIch) {
					woBinIch = woWarIch;
				} else {
					woBinIch = (woBinIch + 30 - 5) % 30;
				}
				maleReglerKonfMenue(rk);
				gebeToolRTippsAus();
				break;
			case KEY_DOWN:
				if (1 == woBinIch) {
					woBinIch = 5;
				} else if (18 == woBinIch) {
					woBinIch = 3;
				} else if (20 == woBinIch) {
					woBinIch = woWarIch;
				} else {
					woBinIch = (woBinIch + 5) % 30;
				}
				maleReglerKonfMenue(rk);
				gebeToolRTippsAus();
				break;
			case (int) '\n':
				switch (woBinIch) {
					case 0:
						/* Regulatorwechsel */
						sicherung = rk;
						while(syncFuR(rk)) {
							eingabe = mvgetch(0, 0);
							switch (eingabe) {
								case KEY_UP:
								case KEY_RIGHT:
									rk = (rk + 1) % 16;
									break;
								case KEY_DOWN:
								case KEY_LEFT:
									rk = (rk + 16 - 1) % 16;
									break;
								case 27:
									rk = sicherung;
									break;
							}
							maleReglerKonfMenue(rk);
							if ('\n' == eingabe || 27 == eingabe) {
								break;
							}
						}
						break;
					case 1:
						/* aktuellen Namen sichern */
						memcpy(name, namen[rk + 32], 20);
						/* Namen fuer die Eingabe loeschen */
						memset(namen[rk + 32], ' ', 20);
						maleReglerKonfMenue(rk);
						wattron(links, COLOR_PAIR(2));
						curs_set(2);
						zu = COLS / 2 - 2;
						for (i = 0; i < 21; i++) {
							zeile = (9 + i) / zu;
							spalte = (9 + i) % zu;
							eingabe = mvwgetch(links, zeile, spalte);
							switch (eingabe) {
								case KEY_UP:
								case KEY_DOWN:
								case KEY_LEFT:
								case KEY_RIGHT:
									i--;
									break;
								case 194:
								case 195:
									eingabe2 = wgetch(links);
									namen[rk + 32][i] = ausZweiMachEins(eingabe,
											eingabe2, zeile, spalte, links);

									break;
								case KEY_BACKSPACE:
								case 127:
									if (0 == i) {
										i--;
									} else {
										i = i - 2;
									}
									mvwaddch(links, zeile, spalte - 1, ' ');
									namen[rk + 32][i - 1] = ' ';
									break;
								case 27:
									break;
								case KEY_RESIZE:
								case ERR:
									sigprocmask(SIG_BLOCK,&mask,&orig_mask);
									if (fensterNeu) {
										passeFensterAn();
										starteMenue();
										for (eingabe2 = (int) i; eingabe2 < 20; eingabe2++) {
											namen[rk + 32][eingabe2] = ' ';
										}
										namen[rk + 32][20] = '\0';
										maleReglerKonfMenue(rk);
										gebeToolRTippsAus();
										wattron(links, COLOR_PAIR(2));
										curs_set(2);
									}
									sigprocmask(SIG_SETMASK,&orig_mask,NULL);
									i--;
									continue;
								default:
									if ((int)'\n' == eingabe) {
										for (eingabe2 = (int) i; eingabe2 < 20; eingabe2++) {
											namen[rk + 32][eingabe2] = ' ';
										}
										namen[rk + 32][20] = '\n';
										break;
									}
									namen[rk + 32][i] = (char) eingabe;
									if (20 == i) {
										mvwaddstr(unten, 0, COLS / 2 - 14,
												"21. Zeichen war kein enter!");
										wrefresh(unten);
										break;
									}
									mvwaddch(links, zeile, spalte, eingabe);
							}
							if (27 == eingabe || (int) '\n' == eingabe) {
								break;
							}
							wrefresh(links);
						}
						if (27 == eingabe) {
							strncpy(namen[rk + 32], name, 20);
							namen[rk + 32][20] = '\0';
						} else if ('\n' == namen[rk + 32][20]) {
							namen[rk + 32][20] = '\0';
							sendeNamen[5] = 0x7E + rk;
							strncpy((char *)&sendeNamen[6], namen[rk + 32], 20);
							pthread_mutex_lock(&mutex3);
							sigprocmask(SIG_BLOCK,&mask,&orig_mask);
							if (-1 == write(fd, sendeNamen, 26)) {
								pthread_mutex_unlock(&mutex3);
								gingNichtGut("write schlug fehl!");
							}
							sigprocmask(SIG_SETMASK,&orig_mask,NULL);
							while (syncP()) {
								if (datenDa) {
									datenDa = FALSE;
									sigprocmask(SIG_BLOCK,&mask,&orig_mask);
									eingabe = read(fd, buf, 8);
									sigprocmask(SIG_SETMASK,&orig_mask,NULL);
									for (i = 0; i < eingabe; i++) {
										buf2[zaehler++] = buf[i];
										fprintf(stderr, "%x ", buf2[zaehler - 1]);
									}
									if (5 < zaehler) {
										mvwprintw(unten, 0, COLS / 2 - 10,
											"Regulatorname %d ", rk + 1);
										if (0x55 == buf2[0] && 0x20 == buf2[1] &&
											0x06 == buf2[2] && 0x50 == buf2[3] &&
											0x01 == buf2[4] && (0x7E + rk) ==
											(unsigned char)buf2[5]) {
											waddstr(unten, "wurde übernommen.");
										} else {
											waddstr(unten,
												"konnte nicht übernommen werden.");
										}
										wrefresh(unten);
										zaehler = 0;
										break;
									}
								}
								usleep(250000);
								if (!datenDa) {
									pthread_mutex_unlock(&mutex3);
									gingNichtGut("Verbindung zum FOM wurde gestört!");
								}
							}
							pthread_mutex_unlock(&mutex3);
						}
						wattroff(links, COLOR_PAIR(2));
						curs_set(0);
						maleReglerKonfMenue(rk);
						break;
					case 2:
						/* Regulator P/I Temperaturuntergrenze */
						*zeiger = rk * 16 + 0x0184;
						helfeglkma(tempN,rk, reglerKonf, 0, 100);
						break;
					case 5:
						/* Regulator Eingabegeraet einstellen */
						*zeiger = rk * 16 + 0x0180;
						helfeglkma(rin,rk, reglerKonf, 0, 16);
						break;
					case 7:
						/* Regulator P/I Temperaturobergrenze */
						*zeiger = rk * 16 + 0x0185;
						helfeglkma(tempH,rk, reglerKonf, 0, 100);
						break;
					case 8:
						/* Regulator Schritt 1 Temperatur */
						*zeiger = rk * 16 + 0x018A;
						helfeglkma(rs1Temp,rk, reglerKonf, 0, 100);
						break;
					case 9:
						/* Regulator Schritt 1 Spannung */
						*zeiger = rk * 16 + 0x0188;
						helfeglkma(rs1U,rk, reglerKonf, 0, 121);
						break;
					case 10:
						/* Regulator Ausgabegeraet einstellen */
						*zeiger = rk * 16 + 0x0181;
						helfeglkma(rout,rk, reglerKonf, 0, 16);
						break;
					case 12:
						/* Regulator P/I Zeit */
						*zeiger = rk * 16 + 0x0186;
						helfeglkma(rZeit,rk, reglerKonf, 1, 120);
					case 13:
						/* Regulator Schritt 2 Temperatur */
						*zeiger = rk * 16 + 0x018C;
						helfeglkma(rs2Temp,rk, reglerKonf, 0, 100);
						break;
					case 14:
						/* Regulator Schritt 2 Spannung */
						*zeiger = rk * 16 + 0x018D;
						helfeglkma(rs2U,rk, reglerKonf, 0, 121);
						break;
					case 15:
						/* Regulator Modus Einstellung */
						*zeiger = rk * 16 + 0x0182;
						helfeglkma(rmod,rk, reglerKonf, 0, 3);
						break;
					case 17:
						/* Regulator P/I Proportinal */
						*zeiger = rk * 16 + 0x0187;
						helfeglkma(rProp,rk, reglerKonf, 0, 121);
					case 18:
						/* Regulator Schritt 3 Temperatur */
						*zeiger = rk * 16 + 0x018E;
						helfeglkma(rs3Temp,rk, reglerKonf, 0, 100);
						break;
					case 19:
						/* Regulator Schritt 3 Spannung */
						*zeiger = rk * 16 + 0x018F;
						helfeglkma(rs3U,rk, reglerKonf, 0, 121);
						break;
					case 20:
						/* Fan off Einstellung */
						*zeiger = rk * 16 + 0x0189;
						helfeglkma(rFo,rk, reglerKonf, 0, 2);
						break;
					case 22:
						/* Regulator P/I Integral */
						*zeiger = rk * 16 + 0x0188;
						helfeglkma(rInt,rk, reglerKonf, 0, 121);
					case 27:
						/* Inverse Output Regelung */
						*zeiger = rk * 16 + 0x0183;
						helfeglkma(rInv,rk, reglerKonf, 0, 2);
						break;
				}
				break;
			/* ESC Taste ab ins obere Menue */
			case 27:
				while(!frageTasteAb());
				pthread_mutex_lock(&mutex2);
				weiter = TRUE;
				pthread_mutex_unlock(&mutex2);
				return;
			/* q Taste: Programm beenden */
			case (int)'q':
				sigprocmask(SIG_BLOCK,&mask,&orig_mask);
				pthread_mutex_lock(&mutex2);
				lauf = FALSE;
				pthread_mutex_unlock(&mutex2);
				sigprocmask(SIG_SETMASK,&orig_mask,NULL);
				return;
		}
	}
}

/* Speichert eine Durchflusseinstellung dauerhaft.
   Bei Erfolg wird True zurueckgegeben beim Scheitern False */
static bool speichereDFM() {
	const unsigned char se[5] = {0x55,0x20,0x05,0x93,0x00};
	/* Auf Antwort vom FOMpro warten */
	if(!verarbeiteAntwort()) {
		return FALSE;
	}
	do {
		/* sende Einstellungen speichern Bytefolge */
		sigprocmask(SIG_BLOCK,&mask,&orig_mask);
		if (-1 == write(fd,se,5)) {
			gingNichtGut("write schlug fehl!");
		}
		sigprocmask(SIG_SETMASK,&orig_mask,NULL);
		
	/* wieder auf Antwort vom FOMpro warten,
	 * wenn es schief ging erneut anfragen */
	} while(!verarbeiteAntwort());
	return TRUE;
}

/* schaut ob das Programm beendet werden muss und passt das Fenster an */
static bool syncFuD() {
	if (!syncP()) {
		return FALSE;
	}
	if (fensterNeu) {
		passeFensterAn();
		starteMenue();
		maleDFMKonfMenue();
		gebeToolDFMTippsAus();
	}
	return TRUE;
}

/* Hilfsfunktion fuer DFMeinstellungen */
static void helfegdkma(unsigned char kanal, unsigned char alarmKonf[7],
		unsigned char minWert, unsigned short maxWert) {
	int eingabe = 0;
	unsigned char sicherung = dfm[kanal];
	while (syncFuD()) {
		eingabe = mvgetch(0, 0);
		switch (eingabe) {
			case KEY_UP:
			case KEY_RIGHT:
				dfm[kanal] = (dfm[kanal] - minWert +  1) % maxWert + minWert;
				break;
			case KEY_DOWN:
			case KEY_LEFT:
				dfm[kanal] = (dfm[kanal] - minWert +  maxWert - 1) % maxWert + minWert;
				break;
			case 27:
				dfm[kanal] = sicherung;
				break;
		}
		maleDFMKonfMenue();
		if ('\n' == eingabe) {
			mvwaddnstr(unten, 0, COLS / 2 - 17, "Einstellung wird gespeichert ...", 33);
			wrefresh(unten);
			alarmKonf[6] = dfm[kanal];
			pthread_mutex_lock(&mutex3);
			do {
				sigprocmask(SIG_BLOCK,&mask,&orig_mask);
				if (-1 == write(fd,alarmKonf,7)) {
					gingNichtGut("write schlug fehl!");
				}
				sigprocmask(SIG_SETMASK,&orig_mask,NULL);
			} while (!speichereDFM());
			mvwaddnstr(unten, 0, COLS / 2 - 17, " Einstellung wurde gespeichert. ", 33);
			wrefresh(unten);
			pthread_mutex_unlock(&mutex3);
			return;
		} else if (27 == eingabe) {
			return;
		}
	}
}

/*Reiter fuer die Konfiguration des Durchflussmessers */
static void gebeDFMKonfMenueAus() {
	int eingabe = 0;
	unsigned short dfmPulse = 0;
	/* Bytefolge um DFMeinstellungen zu setzen */
	unsigned char dfmKonf[7] = {0x55,0x20,0x07,0x91,0x01,0x00,0x00};
	unsigned char sicherung = 0;
	unsigned char sicherung2 = 0;
	woBinIch = 0;

	while(syncFuD()) {
		eingabe = mvgetch(0, 0);
		switch(eingabe) {
			/* Endlich mal ein rechteckiges Menue */
			case KEY_RIGHT:
				woBinIch = (woBinIch + 1) % 2
						+ (woBinIch / 2) * 2;
				maleDFMKonfMenue();
				gebeToolDFMTippsAus();
				break;
			case KEY_LEFT:
				
				woBinIch = (woBinIch + 2 - 1) % 2
						+ (woBinIch / 2) * 2;

				maleDFMKonfMenue();
				gebeToolDFMTippsAus();
				break;
			case KEY_UP:
				woBinIch = (woBinIch + 6 - 2) % 6;
				maleDFMKonfMenue();
				gebeToolDFMTippsAus();
				break;
			case KEY_DOWN:
				woBinIch = (woBinIch + 2) % 6;
				maleDFMKonfMenue();
				gebeToolDFMTippsAus();
				break;
			/* ESC Taste ab ins obere Menue */
			case 27:
				while(!frageTasteAb());
				pthread_mutex_lock(&mutex2);
				weiter = TRUE;
				pthread_mutex_unlock(&mutex2);
				return;
			case (int) '\n':
				switch (woBinIch) {
					case 0:
						/* DFM Alarm Meldung */
						dfmKonf[5] = 0x16;
						helfegdkma(0, dfmKonf, 0, 256);
						break;
					case 1:
						/* DFM Alarm Mainboard */
						dfmKonf[5] = 0x17;
						helfegdkma(1, dfmKonf, 0, 256);
						break;
					case 2:
						/* DFM Alarm Relay 1 */
						dfmKonf[5] = 0x18;
						helfegdkma(2, dfmKonf, 0, 256);
						break;
					case 3:
						/* DFM Alarm Relay 2 */
						dfmKonf[5] = 0x19;
						helfegdkma(3, dfmKonf, 0, 256);
						break;
					case 4:
						/* DFM Alarm 12V Switch */
						dfmKonf[5] = 0x1A;
						helfegdkma(4, dfmKonf, 0, 256);
						break;
					case 5:
						sicherung = dfm[5];
						sicherung2 = dfm[6];
						while (syncFuD()) {
							eingabe = mvgetch(0, 0);
							switch (eingabe) {
								case KEY_UP:
								case KEY_RIGHT:
									dfmPulse = dfm[5] * 256 + dfm[6];
									dfmPulse = (dfmPulse - 10 +  1) % 10000 + 10;
									dfm[5] = (dfmPulse >> 8) & 0x00FF;
									dfm[6] = dfmPulse & 0x00FF;
									break;
								case KEY_DOWN:
								case KEY_LEFT:
									dfmPulse = dfm[5] * 256 + dfm[6];
									dfmPulse = (dfmPulse - 10 +  10000 - 1) % 10000 + 10;
									dfm[5] = (dfmPulse >> 8) & 0x00FF;
									dfm[6] = dfmPulse & 0x00FF;
									break;
								case 27:
									dfm[5] = sicherung;
									dfm[6] = sicherung2;
									break;
							}
							maleDFMKonfMenue();
							if ('\n' == eingabe) {
								mvwaddnstr(unten, 0, COLS / 2 - 17, "Einstellung wird gespeichert ...", 33);
								wrefresh(unten);
								dfmKonf[5] = 0x1C;
								dfmKonf[6] = dfm[6];
								pthread_mutex_lock(&mutex3);
								do {
									sigprocmask(SIG_BLOCK,&mask,&orig_mask);
									if (-1 == write(fd,dfmKonf,7)) {
										gingNichtGut("write schlug fehl!");
									}
									sigprocmask(SIG_SETMASK,&orig_mask,NULL);
								} while (!verarbeiteAntwort());
								dfmKonf[5] = 0x1D;
								dfmKonf[6] = dfm[5];
								do {
									sigprocmask(SIG_BLOCK,&mask,&orig_mask);
									if (-1 == write(fd,dfmKonf,7)) {
										gingNichtGut("write schlug fehl!");
									}
									sigprocmask(SIG_SETMASK,&orig_mask,NULL);
								} while (!speichereDFM());
								mvwaddnstr(unten, 0, COLS / 2 - 17, " Einstellung wurde gespeichert. ", 33);
								wrefresh(unten);
								pthread_mutex_unlock(&mutex3);
								break;
							} else if (27 == eingabe) {
								break;
							}
						}
						break;
				}
				break;
			/* q Taste: Programm beenden */
			case (int)'q':
				sigprocmask(SIG_BLOCK,&mask,&orig_mask);
				pthread_mutex_lock(&mutex2);
				lauf = FALSE;
				pthread_mutex_unlock(&mutex2);
				sigprocmask(SIG_SETMASK,&orig_mask,NULL);
				return;
		}
	}
}

int main(void) {
	/* Auf Default Zeichensetzung stellen */
	if (!setlocale(LC_ALL, "")) {
		fprintf(stderr, "Elende Lokaleinstellungen.\n");
		return EXIT_FAILURE;
	}

	/* Schauen ob utf8 an ist */
	if (strcmp(nl_langinfo(CODESET), "UTF-8")) {
		fprintf(stderr, "Dieses Programm braucht UTF-8!\n");
		return EXIT_FAILURE;
	}

	/* Signale sollten zwar read und write nicht aus dem Konzept bringen
	   und mit Hilfe von EINTR abgefangen werden koennen, aber ich
	   blockiere lieber die Signale und lass write und read in Ruhe arbeiten */         
	sigemptyset(&mask);
	sigaddset(&mask, SIGIO);
	sigaddset(&mask, SIGINT);
	sigaddset(&mask, SIGWINCH);

	/* Versuch mit dem FOM pro Verbindung aufzunehmen */
	oeffnePort();
	testeVerbindung();
	printf("Erstkontakt hergestellt, frage Konfiguration ab...\n");
	holeKonfigurationsdaten();
	holeKonfigurationsdaten2();
	printf("Konfigurationsdaten empfangen, frage Namen ab...\n");
	holeNamen();
	printf("Namen empfangen. Starte ncurses Modus.\n");
	/* Curses Modus starten */
	starteCurses();
	
	while (lauf) {
		fensterNeu = TRUE;
		switch (aktiv) {
			case 0:
				/* Navigation und Einstellungen im Monitormodus */
				if (pthread_create(&thread, NULL, starteMonitorModus, NULL)) {
					gingNichtGut("Konnte keinen Programmfaden anlegen.");
					
				}
				/* Monitordaten des FOM abfragen */
				frageMonitorDatenAb();
				/* Warten bis der frageMonitorModusfaden fertig ist */
				pthread_join(thread, NULL);
				break;
			case 1:
				/* FOM pro Monitor Daten im Hintergrund abfragen */
				if (pthread_create(&thread, NULL, frageFOMDatenAb, NULL)) {
					gingNichtGut("Konnte keinen Programmfaden anlegen.");
					
				}
				/* Luefterkonfigurationsmodus starten */
				gebeLuefterKonfMenueAus();
				/* Warten bis der frageMonitorDatenAbfaden fertig ist */
				pthread_join(thread, NULL);
				break;
			case 2:
				/* FOM pro Monitor Daten im Hintergrund abfragen */
				if (pthread_create(&thread, NULL, frageFOMDatenAb, NULL)) {
					gingNichtGut("Konnte keinen Programmfaden anlegen.");
					
				}
				/* Sensorkonfigurationsmodus starten */
				gebeSensorKonfMenueAus();
				/* Warten bis der frageMonitorDatenAbfaden fertig ist */
				pthread_join(thread, NULL);
				break;
			case 3:
				/* FOM pro Monitor Daten im Hintergrund abfragen */
				if (pthread_create(&thread, NULL, frageFOMDatenAb, NULL)) {
					gingNichtGut("Konnte keinen Programmfaden anlegen.");
					
				}
				/* Reglerkonfigurationsmodus starten */
				gebeReglerKonfMenueAus();
				/* Warten bis der frageMonitorDatenAbfaden fertig ist */
				pthread_join(thread, NULL);
				delwin(mitte);
				break;
			case 4:
				/* FOM pro Monitor Daten im Hintergrund abfragen */
				if (pthread_create(&thread, NULL, frageFOMDatenAb, NULL)) {
					gingNichtGut("Konnte keinen Programmfaden anlegen.");
					
				}
				/* Durchflussmesserkonfigurationsmodus starten */
				gebeDFMKonfMenueAus();
				/* Warten bis der frageMonitorDatenAbfaden fertig ist */
				pthread_join(thread, NULL);
				break;
		}
		weiter = FALSE;
	}
	
	/* Porteinstellungen zurueckschreiben */
	tcsetattr(fd,TCSANOW,&alt);

	/* Serielle Schnittstelle wieder schliessen */
	close(fd);

	/* Fenster schliessen */
	delwin(links);
	delwin(rechts);
	delwin(unten);

	/* curses mode beenden */
	endwin();
	printf("ncurses Modus beendet.\n");
	return EXIT_SUCCESS;
}
